import sys
import os

from khtml import KHTMLPart
from kdecore import KURL
from kdeui import KMainWindow, KPopupMenu, KInputDialog
from qt import QObject, QHBox, QVBox, QSplitter, QListView, QListViewItem, QPushButton, QTimer, QString, SIGNAL, SLOT
import threading
import shelve
import urlparse
import commands
import mimetypes
import urlparse
import datetime
import time
import exceptions
import httplib

_maxEntries = 300

appRoot = os.path.dirname(os.path.abspath(os.path.realpath(sys.argv[0])))

sys.path.append(os.path.join(appRoot, 'thirdparty'))
import feedparser

sys.path.append(os.path.join(appRoot, 'thirdparty/urlgrabber-2.9.9'))
import urlgrabber

_configRoot = os.path.expandvars('$HOME/.KatchTV')
try:
	os.mkdir(_configRoot)
except:
	pass

_mediaStoreRoot = os.path.expandvars('$HOME/.KatchTV/mediastore')
try:
	os.mkdir(_mediaStoreRoot)
except:
	pass
	
_configFile = os.path.join(_configRoot, "datastore.shelf")
_helpContents = "file://localhost" + os.path.join(appRoot, 'Docs/Manual/index.html')
_progressBarImageURI = 'file://localhost' + os.path.join(appRoot, 'webstyle/progressbar.png')

def _mimetypeForHTTPURL(url):
	PROTO=0
	HOSTANDPORT=1
	PATH=2
	
	urlparts = urlparse.urlparse(url)
	
	if urlparts[PROTO] != 'http' and urlparts[PROTO] != 'https':
		# can't figure this out with httplib, but it probably
		# won't be a feed anyway; just return the default
		return 'application/octet-stream'
	
	if ':' in urlparts[HOSTANDPORT]:
		host = urlparts[HOSTANDPORT][:urlparts[HOSTANDPORT].index(':')]
		port = urlparts[HOSTANDPORT][urlparts[HOSTANDPORT].index(':'):]
	else:
		host = urlparts[HOSTANDPORT]
		port = 80
	
	conn = httplib.HTTPConnection(host, port)
	conn.request('HEAD', urlparts[PATH])
	resp = conn.getresponse()
	return resp.msg.gettype()

def _isAFeed(url):
	feedGiveaways = [
		'http://feeds.feedburner.com/',
		'.libsyn.com/',
		'.rss0.99',
		'.rss2.0',
		'.rss2',
		'.rss',
		'rss.php',
		'.rdf',
		'.atom',
		'feed=' ]
		
	feedMimetypes = [
		'application/atom+xml',
		'application/rdf+xml',
		'application/rss+xml',
		'application/x-netcdf',
		'application/xml',
		'text/xml'
	]
	# iTunes seems to have been deliberately broken so this doesn't
	# work any more :(  Funny how people can publish open content
	# on iTunes, and Apple will prevent others from openly recieving
	# it, eh?
	#
	#if url[:len('itms:')] == 'itms:':
	#	# definitely a feed; itunes
	#	newURL = url[len('itms:')+1:]
	#	return True, newURL
	
	assert url.__class__ is str
	
	# handle feed: uris
	prefix = url[:len('feed:')]
	if prefix == 'feed:':
		altPrefix = url[:len('feed://')]
		if altPrefix == 'feed://':
			newURL = url[len('feed://'):]
		else:
			newURL = url[len('feed:'):]
		if not '://' in newURL:
			newURL = 'http://' + newURL
		return True, newURL
	
	# handle rss: uris
	prefix = url[:len('rss:')]
	if prefix == 'rss:':
		altPrefix = url[:len('rss://')]
		if altPrefix == 'rss://':
			newURL = url[len('rss://'):]
		else:
			newURL = url[len('rss:'):]
		if not '://' in newURL:
			newURL = 'http://' + newURL
		return True, newURL
	
	# quick checks for other stuff that looks like a feed; we can
	# afford to be a little optimistic here, since the browser IS
	# a feed browser, and it's likely to be used for picking feeds
	# still need to do this properly by detecting mimetypes and
	# such though.
	for i in feedGiveaways:
		try:
			if i in url:
				return True, url
		except TypeError:
			print "Warning: TypeError in isAFeed.  URL is", url
			return False, None
	
	# no dice; do it the hard (and equally unreliable so far) way
	try:
		mimetype = _mimetypeForHTTPURL(url)
		if mimetype in feedMimetypes:
			return True, url
	except httplib.HTTPException:
		return False, None
	
	# if we get to here, and didn't recognise it, it's probably not a feed
	return False, None

class DownloadThread(threading.Thread):
	class ProgressUpdater:
		def __init__(self, pendingQueueEntry):
			self._pendingQueueEntry = pendingQueueEntry
		
		def start(self, filename=None, url=None, basename=None, length=None, text=None):
			self._pendingQueueEntry['length'] = length
			self._pendingQueueEntry['bytesread'] = 0
			self._pendingQueueEntry['percent'] = 0.0
			self._pendingQueueEntry['finished'] = False
		
		def update(self, bytesread, now=None):
			self._pendingQueueEntry['bytesread'] = bytesread
			if self._pendingQueueEntry['length'] > 0:
				totalLen = float(self._pendingQueueEntry['length'])
				read = float(self._pendingQueueEntry['bytesread'])
				self._pendingQueueEntry['percent'] = (100.0 / totalLen) * read
			else:
				self._pendingQueueEntry['percent'] = 0.0
		
		def end(self, bytesread=None, now=None):
			self._pendingQueueEntry['percent'] = 100.0
			self._pendingQueueEntry['finished'] = True

	def __init__(self, mainWin):
		threading.Thread.__init__(self)
		self._finished = False
		self._grabber = urlgrabber.grabber.URLGrabber()
		self._mainWin = mainWin
		self._loadCompletedList()
		self._loadPendingQueue()
		self._running = False

	def enable(self):
		self._running = True

	def disable(self):
		self._running = False

	def _statusReport(self, msg):
		self._mainWin.statusBar().message(msg)
	
	def _loadCompletedList(self):
		global _configFile
		store = shelve.open(_configFile)
		if store.has_key('completedFiles'):
			self._doneQueue = store['completedFiles']
		else:
			self._doneQueue = {}
		store.close()
	
	def _saveCompletedList(self):
		global _configFile
		store = shelve.open(_configFile)
		store['completedFiles'] = self._doneQueue
		store.close()
	
	def getStatisticsFor(self, uri):
		assert self.haveLocally(uri)
		return self._pendingQueue[uri]

	def _downloadPathInUse(self, fullPath, skipURI=None):
		for uri in self._pendingQueue.keys():
			if uri == skipURI:
				continue
			
			if self._pendingQueue[uri].has_key('localFile'):
				if self._pendingQueue[uri]['localFile'] == fullPath:
					return True
		
		for uri in self._doneQueue.keys():
			if skipURI is not None and uri == skipURI:
				continue
			
			if self._doneQueue[uri]['localFile'] == fullPath:
				return True
		
		return False
	
	def _destPathFromURI(self, uri):
		"""Return the local path where a uri should be stored"""
		urlParts = urlparse.urlparse(uri)
		# take the leading / off the path part, and return it
		pathPart = urlParts[2][1:]
		fullPath = os.path.join(_mediaStoreRoot, urlParts[1], pathPart)
		
		# check if the pathname is already in use, and if it is,
		# append a number that increments until we find a
		# filename that's available.
		pathNameOK = not self._downloadPathInUse(fullPath, skipURI=uri)
		count = 1
		while not pathNameOK:
			# another download has this filename already
			basename = pathPart[:pathPart.rindex('.')]
			exten = pathPart[pathPart.rindex('.'):]
			fullPath = os.path.join(_mediaStoreRoot, urlParts[1], basename + '_' + str(count) + exten)
			pathNameOK = not self._downloadPathInUse(fullPath)
			count += 1
		return fullPath

	def _moveExtensionForMimetype(self, downloadPath, mimetype):
		"""Moves a file to a new filepath, so that its extension matches its mimetype.  Returns the new pathname."""
		# FIXME: not implemented
		return downloadPath

	def _ensureDirExistsFor(self, filePath):
		"""Makes sure that the subdirectory exists for a given filename"""
		fullDirPath = os.path.dirname(filePath)
		try:
			os.makedirs(fullDirPath)
		except OSError:
			pass
	
	def _pruneSubPaths(self, rootPath):
		"""Remove any empty directories under a given root directory path"""
		for root, dirs, files in os.walk(rootPath, topdown=False):
			for dirname in dirs:
				fullPath = os.path.join(root, dirname)
				
				subFiles = os.listdir(fullPath)
				if len(subFiles) == 0:
					os.rmdir(fullPath)
		
		# if we deleted the rootPath (only happens if it's empty), then
		# we re-create it.  Doing it this way may help to free diskspace
		# that isn't always freed by unlinking the files themselves.
		try:
			os.makedirs(rootPath)
		except:
			pass

	def addDownload(self, uri, title, isTorrent=False):
		# CNet encodes the content text as a URL?! Don't
		# want to apply general decoding for a specific
		# case, so just do this
		title = title.replace('%20', ' ')
		
		self._pendingQueue[uri] = {}
		# set the URI as a sub-element too, so the updater has access to it (mostly for debugging)
		self._pendingQueue[uri]['uri'] = uri
		self._pendingQueue[uri]['percent'] = 0.0
		self._pendingQueue[uri]['length'] = 0
		self._pendingQueue[uri]['bytesread'] = 0
		self._pendingQueue[uri]['title'] = title
		# we assume it's not a torrent until we download it and examine the file's actual data
		self._pendingQueue[uri]['isTorrent'] = isTorrent
		
		# We should probably get mimetypes from the feed, at least until we can
		# check the downloaded file's magic.  For now, it's not implemented.
		#self._pendingQueue[uri]['mimetype'] = mimetype
		
		self._pendingQueue[uri]['localFile'] = self._destPathFromURI(uri)
		
		# call this last, since it may depend on previous stuff
		self._pendingQueue[uri]['updater'] = DownloadThread.ProgressUpdater(self._pendingQueue[uri])
		
		# finally, begin the download
		self._pendingQueue[uri]['beginDownload'] = True

	def addCompleted(self, uri, localFile, mimetype, title):
		assert not self._doneQueue.has_key(uri)
		assert uri is not None
		assert localFile is not None
		assert mimetype is not None
		assert title is not None
		
		self._doneQueue[uri] = {}
		self._doneQueue[uri]['localFile'] = localFile
		self._doneQueue[uri]['title'] = title
		self._doneQueue[uri]['mimetype'] = mimetype
		
		currentNumReady = str(self._mainWin.bmarksList.readyToWatchBMark.text(BookmarksListView.COL_NUMNEW))
		if currentNumReady == '':
			currentNumReady = 0
		else:
			currentNumReady = int(currentNumReady)
		self._mainWin.bmarksList.readyToWatchBMark.setText(BookmarksListView.COL_NUMNEW, str(currentNumReady + 1))

	def isDownloading(self, uri):
		return self._pendingQueue.has_key(uri)

	def getCurrentDownloads(self):
		return self._pendingQueue

	def cachedDownloads(self):
		return self._doneQueue
	
	def finish(self):
		self._finished = True

	def purge(self, uri):
		assert self.haveLocally(uri)
		try:
			os.unlink(self._doneQueue[uri]['localFile'])
		except:
			# if it's already gone for some reason, that's OK, just print a warning
			print "Warning: tried to delete '%s', but it was already gone." % self._doneQueue[uri]['localFile']
		del self._doneQueue[uri]
		self._pruneSubPaths(_configRoot)

	def haveLocally(self, uri):
		return self._doneQueue.has_key(uri)
		
	def localPathForURI(self, uri):
		assert self.haveLocally(uri)
		return self._doneQueue[uri]['localFile']

	def titleForLocalFileFromURI(self, uri):
		assert self.haveLocally(uri)
		return self._doneQueue[uri]['title']

	def _getMimetype(self, filepath):
		#NOTE: we use the unix command "file" here, instead of the python mimetypes module, since it understands
		# file contents (file magic) and not just the file's name
		fileOutput = commands.getoutput("file -i '%s'" % filepath)
		if not ':' in fileOutput or not ':' in fileOutput:
			# must be bad
			raise OSError, "Can't figure out the mimetype for %s; the file command failed." % filepath
		else:
			mimetype = fileOutput[fileOutput.index(':')+2:]
			if ';' in mimetype:
				# remove the encoding
				mimetype = mimetype[:mimetypefile.index(';')]
			return mimetype

	def _properExtensionForMimetype(self, mimetype, currentExten):
		"""Finds the correct extension for a mimetype, trying to preserve the current extension, if it's valid."""
		# NOTE: we currently specify strict here, thereby forcing a rename of even slightly dodgy extensions
		# this could be changed, but could make players fail for some types they would otherwise handle.
		validExtens = mimetypes.guess_all_extensions(mimetype, strict=True)
		if currentExten in validExtens:
			return currentExten
		else:
			return mimetypes.guess_extension(mimetype)

	def _renameToReflectMimetype(self, localFile, mimetype):
		currentExten = localFile[localFile.rindex('.'):]
		newExten = self._properExtensionForMimetype(mimetype, currentExten)
		
		if newExten is None:
			return localFile
		
		if currentExten == newExten:
			return localFile
		else:
			newFname = localFile + newExten
			os.rename(localFile, newFname)
			return newFname
	
	def _loadPendingQueue(self):
		store = shelve.open(_configFile)
		if store.has_key('pendingQueue'):
			self._pendingQueue = store['pendingQueue']
			
			# create newProgressUpdater objects for each loaded entry
			for k in self._pendingQueue.keys():
				self._pendingQueue[k]['updater'] = DownloadThread.ProgressUpdater(self._pendingQueue[k])
		else:
			self._pendingQueue = {}
		store.close()

	def _savePendingQueue(self):
		# can't save the DownloadThread.ProgressUpdater
		# objects, so use a temporary queue, and remove
		# them from it before saving
		tempQueue = {}
		tempQueue.update(self._pendingQueue)
		
		for k in tempQueue.keys():
			if tempQueue[k].has_key('updater'):
				del tempQueue[k]['updater']
		
		store = shelve.open(_configFile)
		store['pendingQueue'] = tempQueue
		store.close()

	def run(self):
		while not self._finished:
			time.sleep(1)
			
			if not self._running:
				continue
			
			for uri in self._pendingQueue.keys():
				if not self._pendingQueue[uri].has_key('beginDownload'):
					# this download definition isn't complete, and so isn't ready to download
					continue
				
				if self._pendingQueue[uri]['isTorrent'] == True:
					#self._downloadTorrent(uri, 
					del self._pendingQueue[uri]
					self._statusReport("""%s is a BitTorrent file; can't download it for now""" % self._pendingQueue[uri]['title'])
				else:
					try:
						# if the file already exists, try to resume it, but only once.  If we
						# end up here again, remove the original file and try to reget it
						# from scratch
						if os.path.exists(self._pendingQueue[uri]['localFile']):
							regetMode = 'simple'
						else:
							regetMode = None
						
						self._ensureDirExistsFor(self._pendingQueue[uri]['localFile'])
						try:
							self._grabber.urlgrab(uri, filename=self._pendingQueue[uri]['localFile'], progress_obj=self._pendingQueue[uri]['updater'], reget=regetMode)
						except exceptions.AttributeError:
							return
						self._statusReport("%s has finished; watch it from the Ready to Watch page." % self._pendingQueue[uri]['title'])
					except urlgrabber.grabber.URLGrabError, e:
						if os.path.exists(self._pendingQueue[uri]['localFile']):
							self._statusReport("Couldn't download '%s'.  Removing the previous download attempt, and restarting." % self._pendingQueue[uri]['title'])
							os.unlink(self._pendingQueue[uri]['localFile'])
						else:
							self._statusReport("Couldn't download '%s'.  Removing it from the download list." % self._pendingQueue[uri]['title'])
							del self._pendingQueue[uri]
							continue
					
					# get the downloaded file's mimetype, and rename it accordingly
					mimetype = self._getMimetype(self._pendingQueue[uri]['localFile'])
					self._pendingQueue[uri]['localFile'] = self._renameToReflectMimetype(self._pendingQueue[uri]['localFile'], mimetype)
					
					# FIXME: handle downloaded torrent files here, by putting them back
					# onto the download queue as a torrent file.  Next time around, the
					# torrent downloading part of this if-else statement will then handle it.
					
					# move it to the done queue
					self.addCompleted(uri, self._pendingQueue[uri]['localFile'], mimetype, self._pendingQueue[uri]['title'])
					del self._pendingQueue[uri]

class TimeoutError(Exception): pass

def timelimit(timeout):
    def internal(function):
        def internal2(*args, **kw):
            class Calculator(threading.Thread):
                def __init__(self):
                    threading.Thread.__init__(self)
                    self.result = None
                    self.error = None
                
                def run(self):
                    try:
                        self.result = function(*args, **kw)
                    except:
                        self.error = sys.exc_info()[0]
            
            c = Calculator()
            c.start()
            c.join(timeout)
            if c.isAlive():
                raise TimeoutError
            if c.error:
                raise c.error
            return c.result
        return internal2
    return internal

class FeedCache:
	MAX_SANE_PARSE_SECONDS = 10
	_feedcache = {}

	@timelimit(MAX_SANE_PARSE_SECONDS)
	def _realParse(uriStr):
		return feedparser.parse(uriStr)
	_realParse = staticmethod(_realParse)
	
	def parse(cls, uriStr, forceUpdate=False):
		# FIXME: this doesn't handle refreshes yet!
		if not forceUpdate and cls._feedcache.has_key(uriStr):
			return cls._feedcache[uriStr]
		else:
			startTime = datetime.datetime.utcnow()
			try:
				p = FeedCache._realParse(uriStr)
				cls._feedcache[uriStr] = p
				return p
			except TimeoutError:
				print "Warning feed from %s took more than %s seconds to parse; aborted!" % (uriStr, FeedCache.MAX_SANE_PARSE_SECONDS)
				return None
	parse = classmethod(parse)

class KTVHTMLPart(KHTMLPart):
	def __init__(self, arg, window):
		KHTMLPart.__init__(self, arg)
		self._window = window
	
	def _writePage(self, title, body, otherStyle=None, refreshURL=None, refreshSecs=1, bodyTags="", script=""):
		# this seems to cause pages to disappear after they're written for some weird reason.
		#self.setEncoding('utf8')
		
		str = u'<html><head>'
		
		if refreshURL is not None:
			self.setMetaRefreshEnabled(True)
			str += '<meta http-equiv="refresh" content="%d; URL=%s">' % (refreshSecs, refreshURL)
		
		if title is not None:
			str += '<title>KatchTV - ' + title + '</title>'
		else:
			str += '<title>KatchTV</title>'
		
		if otherStyle is None:
			str += """<style type="text/css">
				BODY {
					background: #5CABE5;
				}
				H1 { color: white; }
				H2 { font-size: -1; }
				DIV.BoxOut {
					background: #85C1F9;
					display: block;
					width: 98%;
					overflow: visible;
					word-wrap: break-word; // for long urls
					align: center;
				}
			</style>"""
		else:
			str += '<style type="text/css">' + otherStyle + '</style>'
			
		if script != "":
			str += '<script language="javascript">' + script + '</script>'
		
		str += '</head><body'
		
		if bodyTags != '':
			str += ' ' + bodyTags
		
		str += '>'
		
		if title is not None:
			str += '<h1>' + title + '</h1>'
			
		str += body + '</body></html>'
		
		self.begin()
		self.write(str)
		self.end()
	
	def _writeUnknown(self):
		self._writePage(u'Unknown Page Requested', u"""<p>An unrecognised page was requested.  You may need to upgrade KatchTV to support this feature.  For now, please return to the <a
		href="katchtv:welcome:">KatchTV Welcome Page</a></p>""")
	
	def _writeWelcome(self):
		self._statusReport('Opening Welcome page...')
		
		body = u"<p>Welcome to KatchTV!</p><p>"
		if len(self._window._downloader._doneQueue) > 0:
			body += u"""There is media <a href="katchtv:ready:">Ready to Watch</a>!  Or, you"""
		else:
			body += u"You"
		body += """ might want to explore the Channel Directories to the side.  Just double-click
		on your chosen directory to load it.  From there, you can subscribe to your favourite shows, just by clicking
		on their feed links."""
		self._writePage(u'KatchTV', body)
	
	def _writeReady(self):
		self._statusReport('Opening Ready To Watch page...')
		
		# mark the ready list's items as viewed
		self._window.bmarksList.readyToWatchBMark.setText(BookmarksListView.COL_NUMNEW, '')
		
		count = 0
		body = u""
		downloadsReady = self._window._downloader.cachedDownloads()
		for dURL in downloadsReady.keys():
			title = downloadsReady[dURL]['title']
			body += u"""
				<div class="BoxOut">
					<h2><a href="katchtv:media:%s:%s">%s</a></h2>
					<a href="katchtv:delete:%s">[delete]</a>
				</div>""" % (self._encodeURL(dURL), self._encodeURL(title), title, self._encodeURL(dURL))
			count += 1
		if count == 0:
			body += u"""<p>Nothing ready to watch so far.  Please
			subscribe to a channel and/or download something.</p>"""
		self._writePage(u'Ready to Watch', body, refreshURL='katchtv:ready:', refreshSecs=1)

	def _writeExpiring(self):
		self._statusReport('Opening Expiring Soon page...')
		self._writePage(u'Expiring Soon', u"""<p>Nothing is
			currently expiring.  This either means that nothing
			has been downloaded, or that all downloads are being kept.</p>""")
	
	def _writeFavourites(self):
		self._statusReport('Opening Favourites page...')
		self._writePage(u'Favourites', u"""<p>You haven't subscribed
			to favourite channels yet.</p>""")

	def _writeDownloads(self):
		global _progressBarImageURI
		
		self._statusReport('Opening Current Downloads page...')
		
		dLoads = self._window._downloader.getCurrentDownloads()
		count = 0
		body = u""
		for dl in dLoads.keys():
			progressBarTotalLen = float(300.0)
			title = dLoads[dl]['title']
			percent = dLoads[dl]['percent']
			size = self._prettyPrintBytes(dLoads[dl]['length'])
			downloaded = self._prettyPrintBytes(dLoads[dl]['bytesread'])
			body += u"""
			<div class="BoxOut">
				<h2>%s</h2>
				<table width="100%%">
					<tr><th width="160" align="left">Progress</th>					<td align="left" bgcolor="black"><img src="%s" width="%s%%" height="22" /></td></tr>
					<tr><th width="160" align="left">Percent complete</th>	<td align="right">%3.1f</td></tr>
					<tr><th width="160" align="left">Total Size</th>					<td align="right">%s</td></tr>
					<tr><th width="160" align="left">downloaded</th>				<td align="right">%s</td></tr>
				</table>
			</div>""" % (title, _progressBarImageURI, percent, percent, size, downloaded)
			count += 1
		
		if count == 0:
			body = u"""<p>Nothing is downloading at the moment.  Please check the <a href="katchtv:ready:">Ready to Watch</a> page, or subscribe to a channel, or manually download something.</p>"""
		
		self._writePage(u'Downloads', body, refreshURL='katchtv:downloads:', refreshSecs=1)

	def _prettyPrintBytes(self, bytes):
		bytes = int(bytes)
		if bytes / 1024:
			# at least 1K
			if bytes / (1024 * 1024):
				# at least 1M
					if bytes / (1024 * 1024 * 1024):
						# at least 1G
						return "%.2f GB" % (bytes / float(1024 * 1024 * 1024))
					else:
						return "%.2f MB" % (bytes / float(1024 * 1024))
			else:
				return "%.2f KB" % (bytes / float(1024))
		else:
			return "%d Bytes" % (bytes)
	
	def _writeFeed(self, feedURI):
		global _maxEntries
		
		self._statusReport('Opening Feed page for %s...' % feedURI)
		
		# mark the feed's articles as viewed, and save the feed list
		bmarkItem = self._window.bmarksList.getItemForURI(feedURI)
		bmarkItem.setText(BookmarksListView.COL_NUMNEW, '')
		self._window.bmarksList._saveBookmarks()
		
		# get the title for the whole feed
		p = FeedCache.parse(feedURI)
		if p is None:
			self._writePage(u'Feed at URI "%s" failed to parse' % feedURI, """<p>Sorry, this feed failed to parse.  The server may have been down, or it may be very slow.  If problems persist with this feed, you may want to remove it your subscribed channels list.""")
			return
		
		if p.feed.has_key('title'):
			feedTitle = p.feed['title'].encode('utf8')
		else:
			feedTitle = feedURI
		
		# begin an empty html body
		body = u""
		
		entryCount = 0
		for ent in p.entries:
			# make sure we don't parse huge feeds for too long
			if entryCount > _maxEntries:
				break
			
			# begin a boxed out area for this entry/episode/article
			body += u"""<div class="BoxOut"><h2>"""
			try:
				title = unicode(ent['title'].encode('utf8'))
			except UnicodeDecodeError:
				print "Unparseable title in feed:", ent['title']
				title = '(unparseable title)'
			body += title.encode('utf8')
			body += """</h2>"""
			
			if ent.has_key('summary'):
				body += u"""<p>%s</p>""" % ent['summary'].encode('utf8')
			
			if ent.has_key('enclosures') and len(ent['enclosures']) > 0:
				body += """<h3>Enclosures</h3>
				<ul>"""
				count = 1
				for enc in ent['enclosures']:
					# mark this enclosure as downloaded if it's already stored locally
					if self._window._downloader.haveLocally(enc['href'].encode('utf8')):
						state = 'Ready to Watch'
					else:
						state = 'Click to Download'
					
					# get a title for the enclosure, if possible, or fudge it
					try:
						entTitle = ent['title'].encode('utf8')
					except UnicodeDecodeError:
						print "Unparseable enclosure title in feed:", ent['title'], "(URI:", feedURI, ")"
						entTitle = "(unparseable enclosure title)"
					title = feedTitle + u': ' + entTitle
					
					# if there's more than one enclosure, then add an incrementing number
					# to the title for each one
					if len(ent['enclosures']) > 1:
						title += ' (enclosure ' + str(count) + ')'
					
					# only print the major type for the mimetype (video, instead of video/mpeg), to
					# make things look nicer
					type = enc.get('type','')
					if '/' in type:
						type = type[:type.index('/')]
					
					# we REALLY don't want this stuff in the type string.  Shouldn't be there
					# since removing the subtype would have removed it, but just in case...
					if ';' in type:
						type = type[:type.index(';')]
					
					# finally, write a listitem for this enclosure
					body += """
					<li>
						<a href="katchtv:media:%s:%s">%s (%s, %s)</a>
					</li>""" % (self._encodeURL(enc['href']), self._encodeURL(title), title, type, state)
					if enc.has_key('length'):
						body += ' [%s]' % self._prettyPrintBytes(enc['length'])
					count += 1
				# finish the list
				body += """</ul>"""
			
			if ent.has_key('link'):
				body += u'<p><a href="'
				body += ent['link'].encode('utf8')
				body += '">Read more...</a></p>'
			
			body += """</div>"""
			
			entryCount += 1
		
		self._writePage(u'Feed "%s"' % feedTitle, body)
	
	def _statusReport(self, msg):
		self._window.statusBar().message(msg)
	
	def _writeMedia(self, url, title):
		if self._window._downloader.haveLocally(url):
			self._statusReport("Playing media %s, please wait a moment while the player loads." % title)
			path = self._window._downloader.localPathForURI(str(url))
			path = "file://localhost" + os.path.abspath(path)
			winWidth = self._window.browser.widget().width() - 40
			winHeight = self._window.browser.widget().height() - 40
#			winWidth = "100%"
#			winHeight = "100%"
			self._writePage(None, """
				<table width="%s" height="%s">
					<tr>
						<td width="%s" height="%s">
							<embed width="%s" height="%s" autostart="true" src="%s" name="player">
								<script language="javascript">
									function onFinished() {
										self.location = 'katchtv:ready:';
									}
								</script>
							</embed>
						</td>
					</tr>
				</table>""" % (winWidth, winHeight, winWidth, winHeight, winWidth, winHeight, path), "BODY { background: rgb(0,0,0) }", script="""
					function focusPlayer() {
						player.focus();
					}
				""",
				bodyTags='onLoad="player.focus();"')
		elif self._window._downloader.isDownloading(url):
			stats = self._window._downloader.getStatisticsFor(url)
			self._writePage(title + ' currently downloading', """
				<p>Sorry, this URI hasn't finished downloading yet.  Currently, %d bytes of out %d total bytes (%3.1f%%) have downloaded.""" \
				% (self._prettyPrintBytes(stats['bytesread']), self._prettyPrintBytes(stats['length']), stats['percent']))
		else:
			self._writePage(title + ' not yet downloaded', """
				<p>This media has not been downloaded yet.  You should <a href="katchtv:add_download:%s:%s">place it on the download queue</a>, so you can watch it later.</p>""" % (self._encodeURL(url), self._encodeURL(title)))

	def _beginAndWriteDownload(self, uri, title):
		self._statusReport('Beginning download %s...' % title)
		
		self._window._downloader.addDownload(uri, title)
		self._writePage(title + ' now downloading.', """
			<p>This media is now downloading.  Please check the <a
			href="katchtv:downloads:">current downloads page</a>
			(on the side panel) for progress reports.</p>
			
			<p>Alternatively, you can just watch something else in
			the meantime.  When the download finishes, the Ready
			to Watch bookmark's "New" column will be updated. 
			The status bar will also inform you that the download
			has completed.</p>""")

	def _deleteLocalFileAndWrite(self, uri):
		title = self._window._downloader.titleForLocalFileFromURI(uri)
		self._window._downloader.purge(uri)
		self._writePage(title + ' deleted', """
			<p>%s has been removed from the local filesystem.  We hope you enjoyed it.</p>
			<p>You may wish to return to the <a href="katchtv:ready:">Ready to Watch</a> page now.</p>""" % title)
		self._statusReport('Local file %s deleted.' % title)

	def _encodeURL(self, uri):
		return uri.replace(':', '__colon__')

	def _decodeURL(self, uri):
		return uri.replace('__colon__', ':')

	def handleCustomURL(self, url):
		page = url[8:].lower()
		if page == 'welcome:':
			self._writeWelcome()
		elif page == 'ready:':
			self._writeReady()
		elif page == 'favourites:':
			self._writeFavourites()
		elif page == 'expiring:':
			self._writeExpiring()
		elif page == 'downloads:':
			self._writeDownloads()
		elif page.startswith('feed:'):
			self._writeFeed(url[8+len('feed:'):].encode('utf8'))
		elif page.startswith('delete:'):
			doomedURL = self._decodeURL(url[8+len('delete:'):].encode('utf8'))
			self._deleteLocalFileAndWrite(doomedURL)
		elif page.startswith('media:'):
			args = url[8+len('media:'):].encode('utf8')
			uri = self._decodeURL(args[:args.index(':')].encode('utf8'))
			title = self._decodeURL(args[args.index(':')+1:].encode('utf8'))
			self._writeMedia(uri, title)
		elif page.startswith('add_download:'):
			args = url[8+len('add_download:'):].encode('utf8')
			uri = self._decodeURL(args[:args.index(':')].encode('utf8'))
			title = self._decodeURL(args[args.index(':')+1:].encode('utf8'))
			self._beginAndWriteDownload(uri, title)
		else:
			self._writeUnknown()
	
	def urlSelected(self, url, button, state, target, args):
		self.goToURL(str(url))

	def goToURL(self, url):
		global _helpContents
		
		# FIXME: should complete partial URL fragments before doing
		# anything else with them

		fullKURL = self.completeURL(url)
		url = str(fullKURL.url())
		
		lowerurl = str(url).lower()
		
		if lowerurl == '':
			return
		
		if lowerurl[:7] == u'mailto:':
			# special case; handle mailto urls, since they're used in the manual for contacting the author with bug reports etc.
			# FIXME: this probably isn't very secure
			os.system(u'kfmclient exec "%s"' % lowerurl)
		elif lowerurl == u"katchtv:help:":
			self.goToURL(_helpContents)
		elif lowerurl[:8] == u'katchtv:':
			self.handleCustomURL(url)
		else:
			isFeed, newUrl = _isAFeed(str(url))
			
			if isFeed:
				if self._window.bmarksList.haveChannel(newUrl):
					self.goToURL(u'katchtv:feed:%s' % newUrl)
				else:
					d = FeedCache.parse(newUrl)
					if d is None:
						self._statusReport("Not adding %s; the feed took too long to parse.  Server down?" % newUrl)
						return
					feedTitle = d.feed.get('title', 'untitled feed')
					self._window.bmarksList.addChannel(newUrl, feedTitle)
			else:
				# before we load this, make sure it's going to be html-like, since
				# khtmlpart won't load other kparts for us
				mimetype = _mimetypeForHTTPURL(url)
				knownHTMLMimetypes = [
					'text/html',
					'text/plain',
					'application/xhtml+xml'
				]
				if not mimetype in knownHTMLMimetypes:
					self._window.statusBar().message("Can't load '%s'; KatchTV only browses webpages and feeds.  Please use an external browser for this." % url)
				else:
					self._window.statusBar().message('Loading URI ' + url)
					self.openURL(fullKURL)

	def dragged(self, ev):
		# if a url is dragged, make sure it's absolute
		# FIXME: not hooked in to signal, and not implemented
		pass

class BookmarksListView(QListView):
	COL_NAME = 0
	COL_NUMNEW = 1
	COL_UPDATE_MINUTES = 2
	COL_URI = 3
	COL_LAST_CHECK_TIMESTAMP = 4

	def __init__(self, parent, win):
		self._window = win
		
		QListView.__init__(self, parent)
		self.setRootIsDecorated(True)
		self.viewport().setAcceptDrops(True)
		self.setAcceptDrops(True)
		self.setShowSortIndicator(True)
		
		self.addColumn('Bookmark')
		self.setSorting(BookmarksListView.COL_NAME,True)
		
		self.addColumn('New')
		self.setSorting(BookmarksListView.COL_NUMNEW,True)
		
		self.addColumn('Update Mins')
		self.setSorting(BookmarksListView.COL_URI,False)
		
		self.addColumn('URI')
		self.setSorting(BookmarksListView.COL_URI,False)
		
		self.addColumn('Last Check Timestamp')
		self.setSorting(BookmarksListView.COL_URI,False)
		self.hideColumn(BookmarksListView.COL_LAST_CHECK_TIMESTAMP)
		
		self._initBookmarks()
		
		QObject.connect(self, SIGNAL('dropped(QDropEvent*)'), self._dropped)
		QObject.connect(self, SIGNAL('itemRenamed(QListViewItem*,int,const QString&)'), self._renamed)
		
		# set a timer to check if channels need updating each minute
		self._updateTimer = QTimer()
		self._updateTimer.connect(self._updateTimer,SIGNAL("timeout()"), self._updateChannels)
		self._updateTimer.start(1000 * 60) # 1000 milliseconds * 60

	def _initBookmarks(self):
		self.dloadsInProgressBMark = self._makeItem(self, name='Downloads in Progress', uri='katchtv:downloads:')
		self.readyToWatchBMark = self._makeItem(self, name='Ready to Watch', uri='katchtv:ready:')

# This stuff isn't implemented yet
#		self.expiringSoonBMark = self._makeItem(self, name='Expiring Soon', uri='katchtv:expiring:')
#		self.favChannelsBMark = self._makeItem(self, name='Favourite Channels', uri='katchtv:favourites:')
#		self.favChannelsBMark.setDropEnabled(True)
		
		self.allChannelsBMark = self._makeItem(self, name='Subscribed Channels')
		self.allChannelsBMark.setDropEnabled(True)
		
		self.directoriesBMark = self._makeItem(self, name='Channel Directories')
#		self.directoriesBMark.setDropEnabled(True)
		self._loadDirectories()
		
		self._deleteQueue = []
		
		# Other, dynamically loaded bookmarks are handled separately
		self._loadBookmarks()

	def _loadDirectories(self):
		bookmarks = {	'https://channelguide.participatoryculture.org/': 'Democracy TV',
#									'http://www.podcast.tv': 'PODCast TV',
#									'http://www.podcastalley.com': 'PODCast Alley (video and audio)'
									'http://www.vodcasts.tv/': 'VODCasts.tv',
									'http://vloglist.com/': 'VLogList',
									'http://www.vlogmap.org/': 'VLogMap',
									'http://zencast.com/channels/videochannels.asp': 'ZENCast: Videos',
									'http://mefeedia.com/feeds/': 'Mefeedia',
									'http://www.realpeoplenetwork.com/': 'RealPeopleNetwork',
									'http://www.vodstock.com': 'VOD Stock',
		}
		for k in bookmarks.keys():
			self.addDirectory(k, bookmarks[k])

	def _loadBookmarks(self):
		global _configFile
		
		# clear the current bookmark list
		item = self.allChannelsBMark.firstChild()
		while item is not None:
			self.takeItem(item)
			nextItem = item.nextSibling()
			del(item)
			item = nextItem
		
		# load the bookmarks
		store = shelve.open(_configFile)
		if store.has_key('bookmarks'):
			bookmarks = store['bookmarks']
		else:
			bookmarks = {
				'http://feeds.feedburner.com/joshleospicks': {
					'name':				"Josh Leo's Picks",
					'updateMins':	60,
					'numNew':			'',
					'lastCheck':	self._encodeDateTime(datetime.datetime(year=datetime.MINYEAR, month=1, day=1))
				}
			}
		store.close()
		
		# now add all loaded/generated bookmarks to the list
		chanList = []
		for uri in bookmarks.keys():
			name = bookmarks[uri]['name']
			numNew = bookmarks[uri]['numNew']
			updateMins = bookmarks[uri]['updateMins']
			lastCheck = bookmarks[uri]['lastCheck']
			channel = self.addChannel(uri=uri, name=name, numNew=numNew, updateMins=updateMins, lastCheck=lastCheck)
			chanList.append(channel)
			
# FIXME: blocking this out for now, because it slows startup.  Make sure updates work OK without.
# the timer should handle it fine though.
#		# update all added channels
#		for channel in chanList:
#			self.updateSingleChannel(channel)

	def _saveBookmarks(self):
		bookmarks = {}
		item = self.allChannelsBMark.firstChild()
		while item is not None:
			if item in self._deleteQueue:
				continue
			
			uri = str(item.text(BookmarksListView.COL_URI))
			bookmarks[uri] = {}
			
			bookmarks[uri]['name'] = str(item.text(BookmarksListView.COL_NAME))
			bookmarks[uri]['updateMins'] = str(item.text(BookmarksListView.COL_UPDATE_MINUTES))
			bookmarks[uri]['numNew'] = str(item.text(BookmarksListView.COL_NUMNEW))
			bookmarks[uri]['lastCheck'] = str(item.text(BookmarksListView.COL_LAST_CHECK_TIMESTAMP))
			
			item = item.nextSibling()
		
		store = shelve.open(_configFile)
		store['bookmarks'] = bookmarks
		store.close()
		self._setStatus('Bookmarks saved.')

	def getItemForURI(self, uri):
		item = self.allChannelsBMark.firstChild()
		while item is not None:
			listedURI = str(item.text(BookmarksListView.COL_URI))
			if listedURI == uri:
				return item
			item = item.nextSibling()
		return None
	
	def _getNewEpisodesFor(self, item, lastCheckDate):
		global _maxEntries
		
		feed = FeedCache.parse(str(item.text(BookmarksListView.COL_URI)), forceUpdate=True)
		if feed is None:
			return 0
		
		newCount = 0
		entryCount = 0
		for e in feed.entries:
			# make sure we don't parse huge feeds for too long
			if entryCount > _maxEntries:
				break
			
			itemdate = datetime.datetime(year=datetime.MINYEAR, month=1, day=1)
			if e.has_key('published_parsed'):
				itemdate = e['published_parsed']
			elif e.has_key('created_parsed'):
				itemdate = e['created_parsed']
			elif e.has_key('updated_parsed'):
				itemdate = e['updated_parsed']
			
			# FIXME: this should be a datetime, or something else; not
			# either/or at this stage.
			if not isinstance(itemdate, datetime.datetime):
				itemdate = itemdate[:6]
				itemdate = datetime.datetime(*itemdate)
			
			if itemdate > lastCheckDate:
				newCount += 1
			
			entryCount += 1
		
		return newCount

	def _encodeDateTime(self, dtime):
		return str(dtime)
	
	def _decodeDateTime(self, dtimeStr):
		# remove extra information if present
		try:
			dtimeStr = dtimeStr[:dtimeStr.rindex('.')]
		except:
			pass
		return datetime.datetime(*(time.strptime(dtimeStr, "%Y-%m-%d %H:%M:%S"))[0:6])

	def updateSingleChannel(self, channel):
		self._setStatus("Updating channel %s" % str(channel.text(BookmarksListView.COL_NAME)))
		
		listedURI = str(channel.text(BookmarksListView.COL_URI))
		updateMinutes = int(str(channel.text(BookmarksListView.COL_UPDATE_MINUTES)))
		lastCheckTime = self._decodeDateTime(str(channel.text(BookmarksListView.COL_LAST_CHECK_TIMESTAMP)))
		currentTime = datetime.datetime.utcnow()
		
		timeDiff = datetime.timedelta(minutes=updateMinutes)
		if (currentTime - lastCheckTime) > timeDiff:
			oldCount = str(channel.text(BookmarksListView.COL_NUMNEW))
			if oldCount == '':
				oldCount = 0
			else:
				oldCount == int(oldCount)
			
			numNew = self._getNewEpisodesFor(channel, lastCheckTime)
			channel.setText(BookmarksListView.COL_LAST_CHECK_TIMESTAMP, str(currentTime))
			if numNew > 0:
				channel.setText(BookmarksListView.COL_NUMNEW,  str(oldCount + numNew))
				self._saveBookmarks()
				# FIXME: it would be nice to bold the listviewitems here, but we need to subclass qlistview for that
	
	def _updateChannels(self):
		channel = self.allChannelsBMark.firstChild()
		while channel is not None:
			self.updateSingleChannel(channel)
			channel = channel.nextSibling()
	
# Favourite Channels aren't implemented yet
#	def addFavouriteChannel(self, title, uri):
#		self._makeItem(self.favChannelsBMark, name=title, uri=uri)

	def _makeItem(self, parent, name, numNew='', updateMinutes='', uri='', lastCheckTimestamp=''):
		# don't check for allChannelsBMark if this is a rootnode -- especially since allChannels might *be* the rootnode ;)
		if parent != self:
			# set different defaults if this is a channel
			if parent == self.allChannelsBMark:
				if updateMinutes == '':
					updateMinutes = 60
				if lastCheckTimestamp == '':
					lastCheckTimestamp = datetime.datetime(year=datetime.MINYEAR, month=1, day=1)
		item = QListViewItem(parent, name, str(numNew), str(updateMinutes), uri, str(lastCheckTimestamp))
		
		# again, don't compare parents if they might not exist yet
		if parent != self:
			# configure item differently if this is a channel or a directory
			if parent == self.allChannelsBMark:
				item.setRenameEnabled(BookmarksListView.COL_UPDATE_MINUTES, True)
			item.setRenameEnabled(BookmarksListView.COL_NAME, True)
		
		return item

	def haveChannel(self, uri):
		return self.getItemForURI(uri) is not None

	def _setStatus(self, msg):
		self._window.statusBar().message(msg)
	
	def addChannel(self, uri, name=None, numNew='', updateMins=60, lastCheck=None):
		if lastCheck is None:
			lastCheck = self._encodeDateTime(datetime.datetime(year=datetime.MINYEAR,month=1,day=1))
		
		if name is not None:
			self._setStatus('Adding bookmark %s' % name)
		else:
			self._setStatus('Adding bookmark for URI %s' % uri)
		
		if self.haveChannel(uri):
			self._setStatus("This URL is already in the bookmarks list.")
		
		if not name:
			# this is a new feed that hasn't been named yet, so try to parse it and get some basic details
			p = FeedCache.parse(str(uri))
			if p is None or p['version'] == '':
				self._setStatus("'%s' doesn't seem like a feed at all. Skipping add." % str(uri))
				return
			else:
				if p.feed.has_key('title'):
					title = p.feed['title']
				else:
					title = defaultTitle
		else:
			title = name
		
		item = self._makeItem(self.allChannelsBMark, name=title, uri=uri, numNew=numNew, updateMinutes=updateMins, lastCheckTimestamp=lastCheck)
		self._setStatus('Bookmark "%s" added' % title)
		
		return item


	def addDirectory(self, uri, title):
		item = self._makeItem(self.directoriesBMark, name=title, uri=uri)

	def _isEditableEntry(self, item):
		if item is None:
			return False
		
		name = item.text(BookmarksListView.COL_NAME)
# Favourite channels and expiring soon aren't implemented yet
#		if item in (self.readyToWatchBMark, self.favChannelsBMark, self.allChannelsBMark, self.expiringSoonBMark, self.dloadsInProgressBMark):
		if item in (self.readyToWatchBMark, self.allChannelsBMark, self.dloadsInProgressBMark):
			return False
		return True
	
	def _renamed(self, item, col, text):
		self._saveBookmarks()
	
	def _dropped(self, qDropEv):
		fmt = 'text/uri-list'
		if qDropEv.provides(fmt):
			qDropEv.accept()
			self.addChannel('new channel', str(qDropEv.encodedData(fmt)))

class KTVMainWindow(KMainWindow):
	def __init__(self, *args):
		self._stopped = True
		apply(KMainWindow.__init__, (self, ) + args)
		self._downloader  = DownloadThread(self)
		self._initWidgets()
		self.browser.goToURL('katchtv:welcome:')
		self._downloader.setDaemon(True)
		self._downloader.start()
		self._stopped = False

	def isStopped(self):
		return self._stopped

	def _initWidgets(self):
		self.statusBar().message("Welcome to KatchTV.  Please wait while the program loads...")
		
		self.mainBox = QHBox(self)
		self.mainSplitter = QSplitter(self.mainBox)
		
		self.bmarksListBox = QVBox(self.mainSplitter)
		self.bmarksListBox.setMaximumWidth(250)
		
		self.bmarksList = BookmarksListView(self.bmarksListBox, self)
		QObject.connect(self.bmarksList, SIGNAL('doubleClicked(QListViewItem *)'), self._bookmarkChosen)
		
		self.browserBox = QVBox(self.mainSplitter)
		self.browser = KTVHTMLPart(self.browserBox, self)
		self.browser.setJScriptEnabled(True)
		QObject.connect(self.browser.browserExtension(), SIGNAL('openURLRequestDelayed(const KURL&, const KParts::URLArgs&)'), self._openURLRequest)
		
		self.setCentralWidget(self.mainBox)
		self.statusBar().message("KatchTV is now ready for use.")
		
		self._buttonBox = QHBox(self.bmarksListBox)
		self._addButton = QPushButton("&Add", self._buttonBox, "")
		QObject.connect(self._addButton, SIGNAL('pressed()'), self._addItem)
		
		self._deleteButton = QPushButton("&Delete", self._buttonBox, "")
		QObject.connect(self._deleteButton, SIGNAL('pressed()'), self._deleteItem)
		
		self._helpButton = QPushButton("&Help", self.bmarksListBox, "")
		QObject.connect(self._helpButton, SIGNAL('pressed()'), self._help)

	def _help(self):
		self.browser.goToURL('katchtv:help:')
	
	def _addItem(self):
		validURI = False
		firstTry = True
		while not validURI:
			if firstTry:
				uri, confirmed = KInputDialog.text("Add Feed Bookmark", "Please enter the channel's feed URI (or URL) here.")
				uri = str(uri)
			else:
				uri, confirmed = KInputDialog.text("Add Feed Bookmark", "The URI you entered was invalid.  Please try again, or cancel.", uri)
				uri = str(uri)
			firstTry = False
			if confirmed:
				validURI = _isAFeed(uri)
			else:
				# user cancelled the input
				return
		self.bmarksList.addChannel(uri)
	
	def _deleteItem(self):
		item = self.bmarksList.currentItem()
		if item.parent() != self.bmarksList.allChannelsBMark:
			self.statusBar().message("This item cannot be deleted.  Only channel bookmarks that you added can be deleted.")
			return
		if item is not None:
			name = item.text(BookmarksListView.COL_NAME)
			if not item.isSelected():
				self.statusBar().message("Please select a channel bookmark to delete.")
				return
			item.parent().takeItem(item)
			del(item)
			self.bmarksList._saveBookmarks()
			self.statusBar().message("%s Deleted.  Bookmarks saved." % name)

	def _openURLRequest(self, kurl, urlargs):
		uri = str(kurl.url())
		self.browser.goToURL(kurl)
	
	def _bookmarkChosen(self, item):
		title = item.text(BookmarksListView.COL_NAME)
		uri = item.text(BookmarksListView.COL_URI)
		if title != '' and uri != '':
			self.browser.goToURL(uri)

	def doCleanup(self):
		self._stopDownloads()
	
	def _stopDownloads(self):
		if not self._stopped:
			self._downloader.finish()
			self._downloader._saveCompletedList()
			self._downloader._savePendingQueue()
			self._downloader = None
			self._stopped = True
